/*
 * This program is Copyright 1998 by Henry Cohen commercial rights are reserved.
 * The program is provided as is with no warranties of any sort including 
 * fitness for any purpose.  In no case will the author be held liable for
 * damages related to the use of this program.

 * The author hereby grants rights to use, copy, modify and create derived
 * products for individual, non-commercial purposes provided a copy of this
 * notice is included.  Any commercial use requires the written approval of 
 * the author.
 * Henry Cohen 1998.
 */

/*
 * Japanese/English dictionary in Java.
 * 100% pure Java. At least no native methods and all Unicode
 * internally we will see about being certified when the time comes.  
 * The dictionary is probably EUCJIS but we can take
 * other encodings as well.  Internally all strings are Java Unicode.
 *
 * I may occasionally make comments about the algorithms or language
 * that are personal opinions rather than illuminations of the code.
 * on those occasions I will stand on a soapbox [soap] (my acknowledgements to
 * Marshall Rose for the rhetorical device.)
 */
package jdic;
import java.io.*;
import java.util.*;
import java.util.zip.*;


/**
 * Dictionary Class - abstract Base class for all dictionaries   
 * The method kstrcmp is defined as an abstract method and must be 
 * overridden in subclasses.  This allows me to have one version that does
 * kana folding and another that does not.  Kana folding is useful for general
 * dictionaries but it is not wanted for the kanji dictionary.
 * The user interface should support multiple dictionaries.  At a minimum
 * users will expect two, Edict and Kanjidict.  The object oriented nature
 * of this program should make it easy to support multiple dictionaries.
 *
 * When instantiated the Dictionary class will open a dictionary file and
 * associated index file.  The dictionary data is read in as a single string
 * and a checksum is computed for the dictionary.  Next the index is read in
 * and the dictionary checksum stored with the index is checked against the
 * one computed for the dictionary.  If they do not match the user is prompted
 * to see if we should generate a new index.  If the user does not want to 
 * generate a new index now we throw a IndexOutOfDateException.  The class
 * also provides methods to search the dictionary.
 *
 *@author Hank Cohen
 *@version Alpha 1
 */ 
abstract class Dictionary {
    /*
     * The comparison methods of the class String are useless for my purposes
     * so just use a char array.
     */
    private char[] dictionary;  
    private int[] index;
    private long indexChecksum, dictionaryChecksum;
    private static SearchList excludedList;
    Runtime runtime;
    private static int BUFF_SIZE = 8192;
    
    /**
     * Constructor for class Dictionary 
     * Read in the dictionary from dictFile and an associated index from dictFile.jdx
     * The parameter makeIndex determines whether the constructor will attempt to
     * recompute an index if none exists or if the existing one's checksum fails
     * to match the dictionary.  If makeIndex is false the constructor will throw
     * an IndexChecksumException, cleanup and return.
     * the encoding parameter specifies the dictionary file encoding method.  All
     * current dictionaries are EUCJIS but we could handle other formats in the
     * future.  In particular using SJIS would make editing the dictionary files
     * easier on a PC or Macintosh.
     *
     *@param dictFile a File object for the dictionary file. We assume that the
     * file exists and is readable.  Exceptions caught in I/O are fatal.
     *@param makeIndex if true make a new index if the checksum fails otherwise
     * throw an exception.
     *@param encoding input file encoding method.  EUCJIS for all existing
     * dictionaries.
     *@exception BadIndexException
     */
    Dictionary (File dictFile, boolean makeIndex, String encoding) 
	 throws BadIndexException, IOException, FileNotFoundException, 
	     SecurityException, UnsupportedEncodingException, EOFException {

	char temp_dbuff[];
	FileInputStream instream;
	Reader translated_input = null;
	DataInputStream data_input = null;
	int nchars = 0;
	StringBuffer indexFilePath;
	Adler32 checksum = new Adler32();
	File indexFile;
	JJDic.mainGui.setStatus("Reading "+dictFile.getName());
	

       runtime = Runtime.getRuntime();
	
	/**
	 * Build an input stream for the dictionary.  The input stream has several
	 * stages.  We need file input code translation and checksum. Buffering
	 * is not necessary because we will do only one read of the entire file
	 * and in that case the underlying stream I/O should be efficient.
	 * Unfortunately the checked inputstream object in the java.util.zip package
	 * is only defined for byte streams.  If we checked the byte stream it would
	 * discover changes between the dictionary and index but it would not be
	 * able to discover a change in the encoding from the external format to
	 * Unicode.  Since I am not certain that the encoding rules are fixed for
	 * all time I want to put the checksum at the tail of the input stream.
	 * The solution is to use a character stream and do the checksum after
	 * the dictionary has been all read in.
	 */
	try {
	    instream = new FileInputStream(dictFile);
	} catch (FileNotFoundException e) {
	    System.err.println("File " + dictFile + " not found.");
	    throw e;
	} catch (IOException e) {
	    System.err.println("IOException on instream"+e);
	    throw e;
	}
	/*
	 * Wrap InputStream with a byte to character translator
	 * and a buffering object.
	 */
	if (encoding == null){
	    /*
	     * Use standard encoding from the environment
	     * Regarding buffer sizes for this buffered reader.  When reading
	     * a dictionary from CD-ROM the default is best because it keeps
	     * the drive spinning. 
	     */
	    translated_input = new BufferedReader( new InputStreamReader(instream), BUFF_SIZE);
	} else if (encoding.equalsIgnoreCase("unicode")) {
	    data_input = new DataInputStream(new BufferedInputStream(instream, BUFF_SIZE));
	}else {	    
	    try {
		translated_input = new BufferedReader(new InputStreamReader(instream, encoding));
	    } catch (UnsupportedEncodingException e){
		System.err.println("Encoding " + encoding + " is unsupported");
		throw e;
	    } 
	}
	/*
	 * Now read in the dictionary and compute a checksum
	 */
	temp_dbuff = new char[instream.available()*2]; //bigger than we need
	if (encoding.equalsIgnoreCase("Unicode")) {
 	    try {
 		for(char i = data_input.readChar(); i != -1; i = data_input.readChar()){
 		    checksum.update((i>>>8) & 0xff);
 		    checksum.update(i & 0xff);
 		    temp_dbuff[nchars++] = i;
 		}
 	    }
 	    catch (EOFException e){
 		// ignore
 	    }
	} else {
	    nchars = translated_input.read(temp_dbuff, 0, temp_dbuff.length);
	    int c;
	    
	    for(int j = 0; j < nchars; j++){
		c = temp_dbuff[j];
		checksum.update((c>>>8) & 0xff);
		checksum.update(c & 0xff);
	    }
	}
	try {
	    dictionary = new char[nchars];
	}
	catch (OutOfMemoryError e){
	    System.err.println("Out of memory. Total "+runtime.totalMemory()+
			       " Free: "+runtime.freeMemory());
	}
	for (int i = 0; i< nchars; i++) 
	    dictionary[i] = temp_dbuff[i];
	
	scan_end = dictionary.length;  // If we need to generate an index we will need this
	dictionaryChecksum = checksum.getValue();
	checksum = null;      // enable garbage collection
	JJDic.mainGui.setStatus(dictFile.getName()+" Checksum: "+dictionaryChecksum);
	
	/*
	 * Free up resources. Close the input stream, make the reference to 
	 * temp_debuff null and run the garbage collector.  This should free
	 * several megabytes of memory.
	 */
	if (encoding.equalsIgnoreCase("Unicode")) {
	    data_input.close();
	    data_input = null;
	} else {
	    translated_input.close();
	    translated_input = null;
	}
	
	temp_dbuff = null;
	
	/*
	 * Next get the index.  Check to see if the file exists. If so
	 * then read it in, if not and the makeIndex flag is true then
	 * build a new index.  If the makeIndex flag is false then 
	 * throw an exception.
	 */
	try{
	    indexFilePath = new StringBuffer(dictFile.getCanonicalPath()+".jndx");
	} catch (IOException e) {
	    System.err.println("Pretty strange! No canonical pathname for a file that we have already opened!"); 
	    /* [Soap] Obviously this is here because the Java compiler requires it */
	    throw e;
	}
	try{
	    indexFile = new File(new String(indexFilePath));
	    if (indexFile.exists ()) {
		JJDic.mainGui.setStatus("Read Index. "+indexFile.getName());

		/*
		 * Index file exists so open it and check the index
		 */
		FileInputStream index_fs = new FileInputStream(indexFile);
		DataInputStream index_ds =  new DataInputStream(
								new BufferedInputStream(index_fs, BUFF_SIZE));
		index = new int[((index_fs.available()-8)/4)];
		indexChecksum = index_ds.readLong();

		if (indexChecksum != dictionaryChecksum){
		    // Checksum Failure
		    if (!makeIndex) throw new BadIndexException();
		    buildIndex(indexFile);
		} else {
		    // Checksums match, read the index
		    int i = 0;
		    try{
			for (i = 0; i<index.length; i++){
			    index[i] = index_ds.readInt();
			}
		    }catch (EOFException e){
			// Unexpected EOF
			System.err.println("Unexpected EOF on Index file. "+
					   i+ "items read. Expected "+ 
					   index.length); 
			if (makeIndex) {
			    buildIndex(indexFile);
			} else {
			    throw new BadIndexException();
			}
		    }
		}
	    } else if (makeIndex) {      // No index file
			JJDic.mainGui.setStatus("Build Index. "+dictFile.getName());
			buildIndex(indexFile);   // Build
	    } else {
		throw new BadIndexException(); // or barf
	    }
	} catch (SecurityException e) {
	    System.err.println("Security Exception on Index File "+e);
	    throw e;
	}
	JJDic.mainGui.setStatus("Index done. "+dictFile.getName());
    }
    /**
     * method buildIndex(File indexFile) build an index for the dictionary.
     * Scan the dictionary string looking for "words" to index.  A word in
     * this case is any kanji or any string of kana or any alphabetic string
     * of greater than 3 characters.  When we find a word we save the starting
     * index in the temporary index table. The temporary index will grow to 
     * accomodate any dictionary. The temporary index is sorted and then
     * the instance variable "index" is allocated to be exactly large enough
     * to hold the index table.  The temp index is then copied to the instance
     * variable.
     * Finally we try to write the index out to a new file but if that fails
     * we swallow the exception and continue.
     */
    private  void buildIndex(File indexFile){
	int[] tmp_index = new int[(dictionary.length*3)/4];
	int count = 0, offset = 0;
	String exclude;
	DataOutputStream indxOut;
	JJDic.mainGui.setStatus(indexFile.getName()+" Build Index; scan");	       	
	/* 
	 * build the exclude list before building the index
	 */
	excludedList = new SearchList(ExcludeList.excludeList[0]);
	for (int i = 1; i< ExcludeList.excludeList.length; i++){
	    excludedList.insert(ExcludeList.excludeList[i]);	    
	}
	
	for (offset = this.scan(); offset != -1; offset = this.scan()){
	    if (count > tmp_index.length) tmp_index = reAllocate(tmp_index);
	    tmp_index[count++] = offset;
	}	
	index = new int[count];
	for (int i = 0; i<count; i++) index[i] = tmp_index[i];
	tmp_index = null;

	JJDic.mainGui.setStatus(indexFile.getName()+" Build Index; sort "+count+" items" );	       	
	this.hsort(index);

	/*
	 * Write out the index file if we can.
	 */
	JJDic.mainGui.setStatus(indexFile.getName()+" Build Index; dump");	       	
	try{
	    indxOut = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(indexFile)));
	} catch (IOException e){
	    return;
	}
	try {
	    indxOut.writeLong(dictionaryChecksum);
	    for (int i = 0; i< index.length; i++) {
		indxOut.writeInt(index[i]);
	    }
	    indxOut.flush();
	} catch (IOException e){
	    return;
	} finally {
	    try {
		indxOut.close();
	    } catch (IOException e) {
		// Ignore
	    }
	}
	JJDic.mainGui.setStatus(indexFile.getName()+" Build Index; done");	       	
    }
    /**
     * Method reAllocate(int[] array): allocate a bigger array and copy old to new.
     * old arrray becomes garbage after the assignment in the caller.
     */ 
    private static int[] reAllocate(int[] iarray) {
	int[] t = new int[(iarray.length*3)/2];
	for(int i = 0; i<iarray.length; i++) t[i] = iarray[i];
	return t;
    }
    /**
     * method hsort() heap sort the dictionary keys.
     * Comparisons are made on the strings in the dictionary at the location
     * indicated by the index.  If the keys are out of order the indices are
     * swapped.  The final result is an index array that orders all of the 
     * words in the dictionary lexicographically.  The entries of the dictionary
     * are already lexicographically ordered by their Japanese meanings
     * this apparently causes lots of trouble for quicksort. All of the Japanese
     * keys are already sorted and all Japanese keys are greater than all 
     * English keys.  Together these facts seem to imply that the entropy of
     * the file is already quite high.  In cases like this Quicksort does 
     * badly.  We don't seem as bad as the O(N**2) that is the worst case
     * but quicksort takes 22 minutes to sort 146,946 keys in kanjidic,
     * heap sort takes only 2 minutes.
     * My next requirement for the sort would be that it be stable; i.e.
     * that keys of equal value should retain their relative positions in
     * the sorted index.  Quicksort and Heapsort are both unstable.  If
     * I can find an O(N log N) sort that is stable I will use it but the
     * only one that I am aware of is merge sort ant that requires extra
     * space equal to the size of the original file and given the tight
     * memory situation already I am unwilling to use a method that requires
     * extra space.
     */
 
    private  void hsort (int[] keys) {
	int N = keys.length;
	int t;
	
	for (int i = N/2; i>=1; i--){
	    this.adjust(keys, i,N);
	}
	for (int i = N-1; i>=1; i--){
	    t = keys[i];
	    keys[i] = keys[0];
	    keys[0] = t;
	    this.adjust(keys, 1, i);
	}
    }
    private void adjust(int[] keys, int i, int N){
	int t = keys[i-1];
	int j = 2*i;

	
	while (j <= N){
	    
	    if (j<N && 
		this.kstrcmp(this.dictionary,keys[j-1], this.dictionary,keys[j],20) < 0) j++;
	    
	    if (this.kstrcmp(this.dictionary, t, this.dictionary,keys[j-1],20) >= 0) {
		keys[(j/2)-1] = t;
		return;
	    }
	    keys[(j/2)-1] = keys[j-1];
	    j = 2*j;
	}
	keys[(j/2)-1] = t;
    }

    /**
     * Abstract method kstrcmp returns -, 0 or + as dict1[i1] is < = or > than dict2[i2].
     * the comparison is terminated by an end of line, punctuation or white space
     * in either substring. 
     * Of course dict1 and dict2 can be the same.
     * We assume that the strings contain at least 1 valid character.
     * This method is abstract so that it can be implemented to fold or not fold
     * kana and case.
     */
    abstract int kstrcmp (char[] dict1, int ix1,char[] dict2, int ix2, int n);
    
    /**
     * method scan() Lexical analyzer for dictionaries.
     * scan the current dictionary and return the offsets of the "words" encountered
     * There is persistent state maintained for the current position in the scan.
     * multiple calls will return subsequent indices until the end of the 
     * dictionary string is reached.  Once the end of the dictionary is reached
     * subsequent calls will return -1.  I have not decided whether or not to
     * include a reset function to restart the scan from the beginning of the
     * dicitonary.  I would like this to be completely self contained but since
     * I need to do multiple dictionaries I will need instance variables to hold
     * the state for each dictionary.
     */
    private int scan_next = 0;
    private int scan_end;    // initialized when the dictionary is read in.
        
    private int scan() {
	int current_position = scan_next;
	int mark_position;
	scan_end = this.dictionary.length;
	
	mainloop: while (current_position < scan_end) {
	    /*
	     * Skip any leading punctuation or whitespace
	     */
	    while (!MojiBunkai.isDictChar(this.dictionary[current_position])) {
		current_position++;
		if (current_position == scan_end) break mainloop;
	    }
	    /*
	     * index all kanji
	     */
	    if (MojiBunkai.isKanji(this.dictionary[current_position])) {
		scan_next = current_position+1;
		return current_position;
	    }
	    /*
	     * Kana word?
	     */
	    if (MojiBunkai.isKana(this.dictionary[current_position])) {
		mark_position = current_position;
		/*
		 *[soap] This is a very nice cup o' Java!
		 * You don't have to test for the end of the array.  Just let
		 * the exception occur and catch it.  Easy!
		 */
		try {
		    while (MojiBunkai.isKana(this.dictionary[++current_position]));
		} catch (ArrayIndexOutOfBoundsException e) {
		    // You don't even have to do anything special here!
		}
		scan_next = current_position;
		return mark_position;
	    }
	    /*
	     * Romaji word - includes numerals
	     */
	    if (Character.isLetterOrDigit(this.dictionary[current_position])){
		mark_position = current_position;
		try {
		    while (Character.isLetterOrDigit(this.dictionary[++current_position]));
		} catch (ArrayIndexOutOfBoundsException e) {
		}
		scan_next = current_position;
		/*
		 * Now we have a word dictionary[mark_position to current_position-1]
		 * check for excluded words and return mark_position if OK.  If it
		 * is an excluded word then continue mainloop.  [Soap] Personally I find
		 * it a bit offensive to my sense of style to generate garbage with
		 * every word in the dictionary but that's Java for you.  There is
		 * no convienient way to slice an array and I want the excludedlist
		 * class to have string arguments for generality.  On the other hand
		 * Java's labeled continue and break are very nice, they eliminate the
		 * need for goto's without introducing any extraneous state variables.
		 */
		if (excludedList.contains(new String(this.dictionary, mark_position, current_position - mark_position))) {
		    continue mainloop;  // excluded word so scan again
		} else {
		    scan_next = current_position;
		    return mark_position;
		}
	    } // Romaji word
	} // mainloop
	/*
	 * If we get here it is because we reached the end of the dictionary
	 */
	return -1;
    }

    /*
     * That's it for initialization functions.  Next search functions.
     */
    /**
     * Method lookup(String key) search the dictionary for all occurrences of the
     * specified key.  returns an array of strings containing all entry lines
     * containing the specified key.  There are 5 steps
     * 1. Binary search for key. This does not necessarily find the first entry so
     * 2. Back up to the first index that matches key
     * 3. Walk forward collecting entries containing key.
     * 4. Package up results as an array of strings
     * 5. Return the results arrray.  We should return the only reference so
     * that once the client is finished the array can be garbage collected.
     * We support keys up to 128 characters long.
     * If the boolean parameter prefix is true then only match if the key is
     * a prefix substring of the dictionary entry.
     */
    public String[] lookup(String key, boolean prefix){
	char[] keychars = new char[128];
	int found, first, last, matches = 0;
	String [] results;
	BitSet isprefix;

	//unpack string to array
	key.getChars(0, key.length(), keychars,0);
	keychars[key.length()] = '\0';

	// binary search for key in index
	found = this.binSearch(keychars, key.length());
	
	if (found > 0) {
	    last = found;
	    // back up to first entry matching key
	    first = found;
	    while (this.kstrcmp(this.dictionary, index[first], keychars, 0, key.length()) == 0) first--;
	    first++;    
	    // scan ahead to last matching entry
	    last = found;
	    while (this.kstrcmp(this.dictionary, index[last], keychars, 0, key.length()) == 0) last++;
	} else {
	    // key not found in dictionary
	    return (String[])null;
	}
	/*
	 * Now first points to the first index matching key and last points to 
	 * one past the last index  matching key.  If we are doing a prefix
	 * match then allocate a bitset with one bit per matching index.
	 * Each matching index is tested to see if it is the start of the line
	 * if so a bit is set
	 * we allocate an array exactly large enough to hold that many lines
	 * and copy the strings into the new array.
	 */
	
	int tail, head;
	if (prefix) {
	    isprefix = new BitSet(last-first);
	    for (int i = 0; i< last-first; i++) {
		if (this.dictionary[index[first+i]-1] == '\n') {
		    isprefix.set(i);
		    matches++;
		} else {
		    isprefix.clear(i);
		}
	    }
	    results = new String[matches];
	    int j = 0;
	    for (int i = 0; i<=last-first; i++) {
		if (isprefix.get(i)) {
		    head = index[first+i];
		    tail = head;
		    while (this.dictionary[tail++] != '\n');
		    results[j++] = new String(this.dictionary, head, tail-head);
		}
	    }
	} else {
	    results = new String[last-first];
	    for (int i = 0; i< results.length; i++){
		head = index[first+i];
		tail = head;
		while (this.dictionary[head-1] != '\n') head--;
	    
		while (this.dictionary[tail++]!= '\n');
		results[i] = new String(this.dictionary, head, tail-head);
	    }
	}
	return results;
    }
    /* 
     * method binSearch(char[] key) binary search in dictionary for key.
     * returns the index of a word that matches key.  This may not be the
     * first such key in the file so we have to do a linear probe back to
     * find the first one.  If no word in the dictionary matches then return
     * -1.
     */
    private int binSearch(char[] key, int n) {
	int lower = 0, upper = index.length;
	int middle, compare;
	
	while (lower <= upper){
	    middle = (lower + upper)/2;
	    try {
		compare = this.kstrcmp(key, 0, this.dictionary, index[middle], n);
	    }
	    catch (ArrayIndexOutOfBoundsException e) {
		System.err.println(e+" index.length "+ index.length + " middle "+ middle);
		compare = -1;
	    }
	    if (compare < 0) {
		upper = middle - 1;
	    } else if (compare == 0) {
		return middle;
	    } else {
		lower = middle + 1;
	    }
	}
	return -1;
    }
}

